void xapi_printversion(void *d);
void xapi_open(void *d);
void xapi_close(void *d);
void xapi_close_window(void *d);
void xapi_set_videomode(void *d);
void xapi_open_window(void *d);
void xapi_pvideomode(void *d);
void xapi_lvideomodes(void *d);
void xapi_pletterbox(void *d);
void xapi_sletterbox(void *d);
void xapi_pwinpos(void *d);
void xapi_pwinsize(void *d);
void xapi_saspect (void *d);
void xapi_swinsize(void *d);
void xapi_ontop(void *d);
void xapi_fullscreen(void *d);
void xapi_pontop(void *d);
void xapi_pfullscreen(void *d);
void xapi_mousepointer(void *d);
void xapi_poverride(void *d);
void xapi_soverride(void *d);
void xapi_swinpos(void *d);
void xapi_exit(void *d);
void xapi_quit(void *d);
void xapi_pfilename(void *d);
void xapi_pduration(void *d);
void xapi_pframerate(void *d);
void xapi_pframes(void *d);
void xapi_poffset(void *d);
void xapi_ptimescale(void *d);
void xapi_ploop(void *d);
void xapi_pseekmode (void *d);
void xapi_sseekmode (void *d);
void xapi_pmwidth(void *d);
void xapi_pmheight(void *d);
void xapi_soffset(void *d);
void xapi_stimescale(void *d);
void xapi_sloop(void *d);
void xapi_sreverse(void *d);
void xapi_pposition(void *d);
void xapi_psmpte(void *d);
void xapi_seek(void *d);
void xapi_pfps(void *d);
void xapi_sfps(void *d);
void xapi_sframerate(void *d);
void xapi_jack_status(void *d);
void xapi_ltc_status(void *d);
void xapi_open_ltc(void *d);
void xapi_close_ltc(void *d);
void xapi_open_jack(void *d);
void xapi_close_jack(void *d);
void xapi_osd_smpte(void *d);
void xapi_osd_frame(void *d);
void xapi_osd_off(void *d);
void xapi_osd_on(void *d);
void xapi_osd_text(void *d);
void xapi_osd_font(void *d);
void xapi_osd_nobox(void *d);
void xapi_osd_box(void *d);
void xapi_osd_avail(void *d);
void xapi_osd_mode(void *d);
void xapi_posd(void *d);
void xapi_psync(void *d);
void xapi_osd_pos(void *d);
void xapi_midi_status(void *d);
void xapi_smididriver(void *d);
void xapi_open_midi(void *d);
void xapi_reopen_midi(void *d);
void xapi_close_midi(void *d);
void xapi_detect_midi(void *d);
void xapi_pmidilibrary (void *d);
void xapi_pmidiclk(void *d);
void xapi_smidiclk(void *d);
void xapi_pmidisync(void *d);
void xapi_smidisync(void *d);
void xapi_bidir_alloff(void *d);
void xapi_bidir_loop(void *d);
void xapi_bidir_noloop(void *d);
void xapi_bidir_frame(void *d);
void xapi_bidir_noframe(void *d);
void xapi_bidir_settings(void *d);
void xapi_bidir_nosettings(void *d);
void xapi_bidir_keyboard(void *d);
void xapi_bidir_nokeyboard(void *d);
void xapi_ping(void *d);
void xapi_null(void *d);
